#!/usr/bin/perl -w
#
#   Randomly configure the username & team a Folding@Home client should use.
#   Copyright (C) 2007  Sebastiaan Couwenberg <sebastic@xs4all.nl>
#   
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Manage a Folding@Home instance to run each new WU for a randomly picked 
# username & team combination in the configuration file.
#
# It uses Dick Howells qd (Queue Dumper) utility to monitor when the F@H
# client, started with -oneunit, has successfully completed a WU.
#
# The F@H clients client.cfg file will be modified to use a new username & team
# combination. And the F@H client will be restarted with the new configuration
# and the -oneunit argument to process only a single WU and quit.
#

use strict;
use Cwd;
use File::Basename;
use Date::Calc;
use XML::Simple;
use Getopt::Mixed "nextOption";

$|=1; # show what's printed without waiting for the newline

# We should wait for the child process before terminating to prevent zombies
$SIG{CHLD} = \&wait_for_child;

my $directory  = dirname($0) .'/';
my $fahclient  = 'FAH504-Linux.exe';
my @arguments  = qw/ -verbosity 9 -oneunit /;
if($^O eq 'darwin') {
	# The Mac OS X FAH client uses ~/Library/Folding@home by default
	# -local forces the FAH client to use the current working directory instead
	unshift @arguments, '-local';
}

my $logname    = 'random-folder.log';
my $logfile    = $directory.$logname;

my $configfile = './config.xml';

my %cfg = ( self      => 'random-folder.pl',
            version   => '0.7.1',
	   
	    directory => $directory,
	    logfile   => $logfile,
	    fahclient => $fahclient,
	    arguments => \@arguments,
	    cli_args  => '',
	    
	    report    => 0,
	    verbose   => 0
          );

# Get commandline parameters
Getopt::Mixed::init("directory=s
                     client=s
                     arguments=s
		     report
		     verbose
		     help
		     d>directory
		     c>client
		     a>arguments
		     r>report
		     v>verbose
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit 0;
	}
	elsif($option eq 'arguments') {
		$cfg{cli_args} = $value;	
	}

	$cfg{directory}  = $value if($option eq 'directory');
	$cfg{fahclient}  = $value if($option eq 'client');

	$cfg{report}     = 1      if($option eq 'report');
	$cfg{verbose}    = 1      if($option eq 'verbose');
}
Getopt::Mixed::cleanup();


# Create the logfile and rotate it if necessary
$cfg{directory} .= '/' if(substr($cfg{directory}, -1, 1) ne '/');
$cfg{logfile}    = $cfg{directory}.$logname;

if( ( -e $cfg{logfile} && -r $cfg{logfile} && -w $cfg{logfile}) ||
    (!-e $cfg{logfile} && -w $cfg{directory})
) {
	if(-e $cfg{logfile}) {
		# rotate the logfile if it's >= 1 MB
		rotate_logfile();
	}
	else {
		# logfile doesn't exist, create it
		create_logfile();
	}
	
	# we can now print to the logfile
	if($cfg{verbose}) {
		print_date("$cfg{self} v$cfg{version} starting.\n");
		print_date("\n");
	}
}
else {
	print "Error: Cannot read/write logfile: $cfg{logfile}\n";
	print "       Check the permissions!\n";

	exit 1;
}


# Change working directory to FAH client directory 
if(-d $cfg{'directory'} && -x $cfg{'directory'}) {
        print_date("Changing working directory to: $cfg{'directory'}\n") if($cfg{verbose});
        Cwd::chdir $cfg{'directory'};
        print_date("\n") if($cfg{verbose});
}
else {
        print_date("Error: Cannot change to directory: $cfg{'directory'}\n");
        exit 1;
}


# If the arguments is given, generate the report and exit
if($cfg{report}) {
	generate_report();
	exit 0;
}


# Parse the commandline arguments for the FAH client
$cfg{arguments} = parse_arguments($cfg{cli_args});


# Create the username/team accounts list which will keep control how many times
# an account will be in the pool from which to choose a random entry.
#my @account = reset_accounts();
my @account = load_account_pool();


# Start the F@H client
if(-x $cfg{'fahclient'}) {
	# start the client
	my $pid = start_client();
	
	# wait for the client to terminate, reconfigure the client, and restart it
	manage_client($pid);
}
else {
	print_date("Error: Cannot execute FAH client: $cfg{'fahclient'}\n");
	exit 1;
}


exit(0);


#################################[ Subroutines ]################################

sub print_date {
	my $string = shift;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

	my $message = "[$date $time] $string";
	
	print_log($message);
}

sub print_log {
	my $string = shift;
	
	# print to STDOUT
	print $string;
	
	# print to logfile
	open(LOG, ">>", $cfg{logfile}) || die "Error: Cannot open logfile: $cfg{logfile} ($!)";
	print LOG $string;
	close LOG;
}

sub wait_for_child {
	wait;
}

sub normalize_filesize {
	my $size = $_[0] =~ /^\d+(\.\d+)?$/  ? $_[0] : 0;
	my $unit = $_[1] =~ /^[KMGTPEZY]?B$/ ? $_[1] : 'B';

	# Units:
	# B : byte
	# KB: kilobyte
	# MB: megabyte
	# GB: gigabyte
	# TB: terabyte
	# PB: petabyte
	# EB: exabyte
	# ZB: zettabyte
	# YB: yottabyte

	my $i     = 0;
	my @units = qw/B KB MB GB TB PB EB ZB YB/;
	foreach my $u (@units) {
		if($unit =~ /^$u$/i) {
			if($size >= 1024) {
				$size /= 1024;
				$unit  = $units[$i+1];
			}
			else {
				last;
			}
		}
		$i++;
	}

	return ($size, $unit);
}

sub create_logfile {
	open(LOG, ">", $cfg{logfile}) || die "Error: Cannot open logfile: $cfg{logfile} ($!)";
	close LOG;
}

sub move_historic_files {
	my $file = $cfg{logfile};
	
	# keep only 4 historic logfiles
	my $max = 4;

	# move the files up one index
	for(my $i = ($max-1); $i >= 0; $i--) {
		# <filename>.3 -> <filename>.4
		# ...
		# <filename>.0 -> <filename>.1
		# <filename>   -> <filename>.0
		
		my $old   = $file .'.'. $i;
		my $new   = $file .'.'. ($i + 1);
		if(-e $old) {
			rename $old, $new;
		}
	}

	# finally move the current logfile to the first historic logfile
	my $old   = $file;
	my $new   = $file.'.0';
	if(-e $old) {
		rename $old, $new;
	}
}

sub shift_files {
	my $dir = dirname($cfg{logfile}).'/';
	if(-e $dir && -d $dir) {
		#  current logfile: <filename>
		# previous logfile: <filename>.0
		# historic logfile: <filename>.1
		# ...
		# historic logfile: <filename>.4
		
		# first move all the historic files:
		move_historic_files();

		# create the logfile
		create_logfile();
	}
	else {
		print_date("Error: No such directory: $dir\n");
		exit 1;
	}
}

sub rotate_logfile {
	my $verbose = shift;
	my $file    = $cfg{logfile};
	
	print_date("Rotating logfile if required: $file\n") if($verbose);

	if(-r $file && -w dirname($file)) {
		my $size = (lstat($file))[7];
		
		# 1 MB = 1024 KB = 1048576 B
		my $mb = 1024 * 1024;
		
		# convert the filesize to human friendly format
		my ($fsize, $unit) = normalize_filesize($size, 'B');
		if($fsize =~ /^\d+\.\d+$/) {
			$fsize = sprintf("%.2f", $fsize);
		}
		print_date("Current filesize: $fsize $unit\n") if($verbose);
		
		($fsize, $unit) = normalize_filesize($mb, 'B');
		if($fsize =~ /^\d+\.\d+$/) {
			$fsize = sprintf("%.2f", $fsize);
		}
		
		# Rotate the logfile if its filesize >= 1 MB
		if($size >= $mb) {
			print_date("Rotating because size >= $fsize $unit\n") if($verbose);

			# Move the logfile up 1 index, keep 5 logfiles max
			shift_files();
		}
		else {
			print_date("Not rotating because size < $fsize $unit\n") if($verbose);
		}
		print_date("\n") if($verbose);
	}
	else {
		print "Error: Cannot read file and/or write directory: $file\n";
		exit 1;
	}
}

sub parse_arguments {
	# parse and sanitize the F@H client commandline parameters
	my $value = shift;
	
	# The Linux F@H client supports the following arguments:
	#
	#  -config      Configure user information
	#  -configonly  Configure user information, then exit
	#  -help        Dispaly this usage screen
	#  -queueinfo   Get information on queued work units
	#  -delete x    Delete item #x from work queue
	#  -send x      Send result #x to server then exit. Use x=all to send all results
	#  -verbosity x Sets the output level, from 1 to 9 (max). The default is 3
	#  -pause       Pause after finishing & trying to send current unit
	#  -oneunit     Exit after completing one unit
	#  -forceasm    Force core assembly optimizations to be used if available
	#  -advmethods  Use new advanced scientific cores and/or work units if available
	#  -freeBSD     Make brandelf system call on downloaded cores.
	#  -openBSD     Make elf2olf system call on downloaded cores.
	#
	# The Mac OS X F@H client supports the following arguments:
	#
	#  -config      Configure user information
	#  -configonly  Configure user information, then exit
	#  -help        Dispaly this usage screen
	#  -queueinfo   Get information on queued work units
	#  -delete x    Delete item #x from work queue
	#  -send x      Send result #x to server then exit. Use x=all to send all results
	#  -verbosity x Sets the output level, from 1 to 9 (max). The default is 3
	#  -pause       Pause after finishing & trying to send current unit
	#  -oneunit     Exit after completing one unit
	#  -forceasm    Force core assembly optimizations to be used if available
	#  -advmethods  Use new advanced scientific cores and/or work units if available	
	#
	# On Mac OS X (and Windows) this argument is also supported:
	#
	#  -local       Use configuration files from local directory
	#
	# On both Linux and Mac OS X the v6 FAH client supports the argument:
	#  -smp         Request SMP work units
	# 
	# Only these arguments will be accepted:
	#
	#  -forceasm    Force core assembly optimizations to be used if available
	#  -advmethods  Use new advanced scientific cores and/or work units if available
	#  -freeBSD     Make brandelf system call on downloaded cores.
	#  -openBSD     Make elf2olf system call on downloaded cores.
	#  -local       Use configuration files from local directory
	#  -smp         Request SMP work units
	
	print_date("Parsing FAH client commandline arguments.\n") if($cfg{verbose});
	
	my $valid = 1;
	my @arguments = @{$cfg{arguments}};

	foreach(split /\s+/, $value) {
		if(/^-(forceasm|advmethods|freeBSD|openBSD|local|smp)$/) {
			print_date("Adding argument: $_\n") if($cfg{verbose});
			push @arguments, $_;
		}
		else {
			print_date("Error: Invalid argument: $_\n");
			$valid = 0;
			last;
		}
	}

	if($valid) {
		print_date("\n") if($cfg{verbose});
		
		return \@arguments;
	}
	else {
		print_date("Error: Invalid arguments to start FAH client!\n");
		print_date("Supported arguments:\n");
		print_date(" -forceasm\n");
		print_date(" -advmethods\n");
		print_date(" -freeBSD\n");
		print_date(" -openBSD\n");
		print_date(" -local\n");
		print_date(" -smp (v6 FAH client)\n");
		
		exit 1;
	}
}

sub valid_account {
	my $account = shift;

	if(defined $account                &&
	   defined $account->{team}        &&
	   defined $account->{username}    &&
	   defined $account->{weight}      &&
	   $account->{team}     =~ /^\d+$/ &&
	   $account->{username} !~ /^s*$/  &&
	   $account->{weight}   =~ /^\d+$/
	) {
		return 1;
	}
	else {
		return 0;
	}
}

sub save_account_pool {
	# Save the account pool state in config.xml
	
	my $config = XMLin( $configfile,
	                    KeyAttr    => 'id',
			    ForceArray => [ 'account' ],
			    KeepRoot   => 1
			  ) || die $@;

	delete $config->{config}->{state}->{accounts};

	my $id = 0;
	foreach(@account) {
		if(valid_account($_)) {
			$config->{config}->{state}->{accounts}->{account}->{"$id"} = $_;
		
			$id++;
		}
	}

	XMLout( $config,
	        OutputFile => $configfile,
		KeyAttr    => { account => 'id' },
		NoAttr     => 1,
		KeepRoot   => 1,
	      ) || die $@;
}

sub load_account_pool {
	# load the account pool from the saved state in config.xml
	# if there is no saved state, reset the account pool to build
	# it from scratch

	my $config = XMLin( $configfile,
	                    KeyAttr    => 'id',
			    ForceArray => [ 'account' ]
		          ) || die($@);
			   
	my @account;

	if(defined $config->{state}             && 
	   defined $config->{state}->{accounts} && 
	   defined $config->{state}->{accounts}->{account}
	) {
		foreach my $id (sort keys %{$config->{state}->{accounts}->{account}}) {
			if(valid_account($config->{accounts}->{account}->{"$id"})) {
				push @account, $config->{accounts}->{account}->{"$id"};
			}
		}

		if($#account == -1) {
			@account = reset_accounts();
		}
	}
	else {
		@account = reset_accounts();
	}

	return @account;
}

sub reset_accounts {
	# setup the account pool

	my $config = XMLin( $configfile,
			    KeyAttr    => 'id',
			    ForceArray => [ 'account' ]
			  ) || die($@);
	
	my @account;
	
	foreach my $id (sort keys %{$config->{accounts}->{account}}) {
		my $weight = $config->{accounts}->{account}->{"$id"}->{weight};
		
		if($weight && $weight > 0) {
			for(my $i = 0; $i < $weight; $i++) {
				push @account, $config->{accounts}->{account}->{"$id"};
			}
		}
		else {
			push @account, $config->{accounts}->{account}->{"$id"};
		}
	}

	return @account;
}

sub random_account {
	# pick a random account from the account pool and remove the account 
	# from the pool. if the account pool becomes empty, reset it.

	if($cfg{verbose}) {
		print_date("Account pool:\n");
	}
	
	my $id = 0;
	foreach(@account) {
		my $username = $_->{username};
		my $team     = $_->{team};
		my $weight   = $_->{weight};
		
		if($cfg{verbose}) {
			print_date(" ID      : $id\n");
			print_date(" Username: $username\n");
			print_date(" Team    : $team\n");
			print_date(" Weight  : $weight\n");
			print_date("\n");
		}
		
		$id++;
	}
	
	my $random         = int(rand($#account + 1));
	my $random_account = $account[$random];
	   $random_account->{id} = $random;

	# remove the accout from the pool
	if($cfg{verbose}) {
		print_date("Random account: $random_account->{id}\n");
		print_date("      Username: $random_account->{username}\n");
		print_date("          Team: $random_account->{team}\n");
		print_date("\n");
	}
	
	print_date("Removing account from the pool.\n") if($cfg{verbose});
	splice @account, $random, 1;

	if($#account == -1) {
		print_date("Account pool is now empty, resetting...\n") if($cfg{verbose});
		@account = reset_accounts();

		print_date("\n") if($cfg{verbose});

		# Since we've reset the account pool, 
		# lets show the account selection statistics
		#
		# Explicitly disable verbosity during the reporting
		# to not show the logfile parsing messages
		if($cfg{verbose}) {
			$cfg{verbose} = 0;

			generate_report();
		
			$cfg{verbose} = 1;
		}
	}

	save_account_pool();

	print_date("\n") if($cfg{verbose});
	
	return $random_account;
}

sub reconfigure_client {
	# change the F@H client config to use a randomly picked username/team combination
	
	if($cfg{verbose}) {
		print_date("\n");
		print_date("Reconfiguring FAH client.\n");
		print_date("\n");
	}
	
	# Pick a random account
	my $account = random_account();

	# Check the client.cfg file
	my $clientcfg = $cfg{directory}.'client.cfg';
	if(!-e $clientcfg) {
		print_date("Error: FAH client not configured yet!\n");
		print_date("First run: $cfg{fahclient} -configonly\n");
		
		exit 1;
	}

	print_date("Modifying client.cfg\n") if($cfg{verbose});
	if(-r $clientcfg && -w $clientcfg) {
		# modify the username and team number and
	
		my $data = '';
		open(F, $clientcfg) || die("Error: Cannot open file: $clientcfg ($!)\n");
		while(<F>) {
			#username=Anonymous
			if(/^(username=)(.*?)\s*$/) {
				my $key   = $1;
				my $value = $2;
				
				my $line = $_;
				if($line =~ s/^\Q$key$value\E/$key$account->{username}/) {}
				
				$data .= $line;
			}
			#team=0
			elsif(/^(team=)(.*?)\s*$/) {
				my $key  = $1;
				my $team = $2;
				
				my $line = $_;
				if($line =~ s/^\Q$key$team\E/$key$account->{team}/) {}
				
				$data .= $line;
			}
			else {
				$data .= $_;
			}
		}
		close F;
			
		open(F, ">", $clientcfg) || die("Error: Cannot write file: $clientcfg ($!)\n");
		print F $data;
		close F;
		
		print_date("\n") if($cfg{verbose});
	}
	else {
		print_date("Error: Cannot read and/or write $clientcfg\n");
		
		exit 1;
	}
}

sub start_client {
	# run the F@H client

	print_date("Forking...\n\n") if($cfg{verbose});
	my $pid = fork();
	if($pid) {
		# parent process
		sleep 1; # give the child a little time to start up
		
		print_date("Parent PID: $$\n") if($cfg{verbose});
		
		return $pid;
	} 
	elsif($pid == 0) {
		# child process
		print_date("Child PID: $$\n") if($cfg{verbose});	
	
		if($cfg{verbose}) {
			print_date("Starting FAH client: ./$cfg{fahclient}");
			foreach(@{$cfg{arguments}}) {
				print_log(" $_");
			}
			print_log("\n\n");
		}

		# Run the F@H client discarding its output,
		# the clients output is written to FAHlog.txt
		#`./$cfg{fahclient} @{$cfg{arguments}} > /dev/null 2>&1`;
		my $status = system("./$cfg{fahclient} @{$cfg{arguments}} > /dev/null 2>&1");
		
		if($cfg{verbose}) {
			print_log("\n");
			print_date("Child PID: $$\n");
			print_date("FAH client exited with status: $status\n\n");
		}
		
		# quit the child process when the F@H client has quit
		exit 0;
	} 
	else {
		print_date("Error: Cannot fork! ($!)\n");

		exit 1;
	}
}

sub is_running {
	my $process = shift;

	if($process =~ /^\s*$/) {
		print_date("Error: No process specified!\n");
		exit 1;
	}

	# check if the process is running
	print_date("Checking if process is running: $process\n") if($cfg{verbose});
	
	my $cmd    = "ps aux | egrep \"$process\" | egrep -v \"grep|random-folder\" |";
	open(CMD, $cmd) || die("Error: Cannot open pipe: '$cmd' ($!)");
	my @output = <CMD>;
	close CMD;
	
	if($#output >= 0) {
		# one or more instances of the process is running
		print_date("Yes, process is running\n") if($cfg{verbose});

		return 1;
	}
	else {
		print_date("No, process is not running\n") if($cfg{verbose});

		return 0;
	}
}

sub client_has_quit {
	# check if the child process which started the client has quit
	# and if all FAH client processes have quit,
	# when started with -oneunit the client should quit
	# after completing a single work unit
	# 
	# Linux CPU client processes v5: FAH50*-Linux.exe, FahCore_*.exe
	# Linux SMP client processes v5: fah5, mpiexec, FahCore_*.exe
	# Linux SMP client processes v6: fah6, mpiexec, FahCore_*.exe
	
	my $pid = shift;
	
	my $child_running = kill 0, $pid;
	if($cfg{verbose}) {
		print_date("Checking if process is running: child process $pid\n");
		if($child_running) {
		        print_date("Yes, process is running\n");
		}
		else {
			print_date("No, process is not running\n");
		}
	}

	my @processes = ();
	
	# v5.02: FAH502-Linux.exe
	# v5.04: FAH504-Linux.exe
	if($cfg{fahclient} =~ /^FAH5\d{2}-Linux.exe$/) {
		push @processes, ('FAH5.*-Linux.exe', 'FahCore_.*.exe');
	}
	# v5.91: fah5 
	# v6.00: fah6 
	elsif($cfg{fahclient} =~ /^fah(5|6)$/) {
		push @processes, ($cfg{fahclient}, 'mpiexec', 'FahCore_.*.exe');
	}
	else {
		print_date("Error: Unsupported FAH client: $cfg{fahclient}\n");
		
		exit 1;
	}

	# if the client type cannot be determined,
	# assume the client is still running
	if($#processes >= 0) {
		my $running = -1;
		foreach(@processes) {
			if(is_running($_)) {
				$running = 1;
				
				# not all processes have quit
				return 0;
			}
			else {
				$running = 0;
			}
		}

		if($running == 0 && !$child_running) {
			# all processes have quit
			return 1;
		}
	}

	return 0;
}

sub manage_client {
	# wait for the client to terminate, reconfigure it, 
	# and restart the client with the new configuration
	my $pid = shift;

	print_date("Entering client management loop.\n\n") if($cfg{verbose});
	while(1) {
		# indefinetly wait for the client to finish the WU and terminate
		#
		# adjust the sleep time between checks semi-intelligently:
		#  check every 60 minutes if progress <   80%
		#  check every 30 minutes if progress <   95%
		#  check every  5 minutes if progress >=  95%
		#  check every  1 minutes if progress == 100%
		
		# Rotate the logfile if necessary
		rotate_logfile($cfg{verbose});
		
		print_date("Checking WU status & progress ...\n") if($cfg{verbose});
		my %current_wu = get_current_wu_info();
		
		# sleep 60 minutes by default
		my $sleep = 60 * 60;
		
		# <status>folding now</status>
		print_date("Status  : $current_wu{status}\n") if($cfg{verbose});
		
		# <progress>30%</progress>
		print_date("Progress: $current_wu{progress}\n") if($cfg{verbose} && $current_wu{progress});
		if($current_wu{progress} && $current_wu{progress} =~ /^(\d+(\.\d+)?)\%$/) {
			my $progress = $1;

			if($progress == 100) {
				$sleep = 60 * 1;
			}
			if($progress >= 95 && $progress < 100) {
				$sleep = 60 * 5;
			}
			if($progress < 95) {
				$sleep = 60 * 30;
			}
			if($progress < 80) {
				$sleep = 60 * 60;
			}
		}
		
		# check the status of the current WU
		# if it is finished, check if the F@H client, cores 
		# and mpiexec (for the SMP client) have terminated
		# if these processes have quit, reconfigure the client
		# and restart it with the new configuration
		if($current_wu{status} && $current_wu{status} eq 'finished') {
			# check every minute
			$sleep = 60;
			
			print_date("Finished: $current_wu{end_timestamp}\n") if($cfg{verbose} && $current_wu{end_timestamp});

			print_date("WU has finished!\n") if($cfg{verbose});
			        
			# check if the FAH client processes have quit
			if(client_has_quit($pid)) {
			        # Configure a random userame/team combination
			        reconfigure_client();
				
			        # start the client
			        $pid = start_client();
			}
		}
		elsif($current_wu{status} && $current_wu{status} eq 'ready for upload') {
			# check every minute
			$sleep = 60;
			
			print_date("WU is ready for upload.\n") if($cfg{verbose});
		}
		elsif($current_wu{status} && $current_wu{status} eq 'deleted') {
			# check every minute
			$sleep = 60;
			
			print_date("WU was deleted!\n") if($cfg{verbose});
		}
		else {
			print_date("Expected: $current_wu{expect_timestamp}\n") if($cfg{verbose} && $current_wu{expect_timestamp});
			
			print_date("WU is not finished yet.\n") if($cfg{verbose});
		}
	
		print_date("Waiting for ". ($sleep / 60) ." minutes before polling again.\n") if($cfg{verbose});

		sleep $sleep;
		
		print_log("\n") if($cfg{verbose});
	}
}

sub get_current_wu_info {
	# get the current WU information from the qd output
	$SIG{CHLD} = 'IGNORE';

	my $xml = `./qd -x` || die "Error: Failed to execute './qd -x' ($!)\n";

	$SIG{CHLD} = \&wait_for_child;
	
	# parse the XML file
	my $queue_dump = XMLin( $xml,
	                        KeyAttr    => 'number',
			        ForceArray => [ 'index', 'server' ]
	                      ) || die($@);

	if($queue_dump->{error}) {
		my $error = $queue_dump->{error};
		if($error =~ s/&apos;/'/g) {}
		if($error =~ s/&gt;/>/g) {}
		if($error =~ s/&lt;/</g) {}

		print_date("Error: qd output contains an error!\n");
		print_date("$error\n");
	}

	my %current_wu = ();
	if($queue_dump->{queue} && $queue_dump->{queue}->{current_index} >= 0) {
		my $current_index = $queue_dump->{queue}->{current_index};
	
		%current_wu = %{$queue_dump->{queue}->{index}->{"$current_index"}};
	}
	
	return %current_wu;
}

sub generate_report {
	# parse all the random-folder logfiles
	# and report statistics on how many times
	# an account has been selected
		
	# there should only be 4 historic logfiles
	my $max = 4;

	my @files;

	# move the files up one index
	for(my $i = $max; $i >= 0; $i--) {
		push @files, $logname.'.'.$i;
	}
	push @files, $logname;

	my %logfile;
	foreach my $file (@files) {
		if(-e $file) {
			%logfile = parse_logfile($file, %logfile);
		}
	}
		
	if(defined $logfile{selected_accounts}) {
		print_date("\n") if($cfg{verbose});
		
		# Per account stats
		my %account       = ();
		my $longest_key   = 0;
		my $longest_value = 0;
		my $string_length = 0;

		foreach(@{$logfile{selected_accounts}}) {
			my $account  = $_->{account};
			my $username = $_->{username};
			my $team     = $_->{team};
		
			my $key = $username.'#'.$team;
			
			$account{$key} += 1;
			$longest_key    = length($key)           if(length($key)           > $longest_key);
			$longest_value  = length($account{$key}) if(length($account{$key}) > $longest_value);
		}
		$longest_key += 2; # Account for the two braces that will be printed around the team
		
		print_date("Randomly selected account statistics:\n");
		print_date("\n");
		foreach my $key (sort { $account{$b} <=> $account{$a} } keys %account) {
			my ($username, $team) = split /#/, $key;

			my $length_key   = $longest_key   - length($key);
			my $length_value = $longest_value - length($account{$key}) + 1; # Account for the : that will be printed

			my $percentage   = sprintf("%.2f", (100 / ($#{$logfile{selected_accounts}} + 1)) * $account{$key});
			
			# username (team): x times
			my $string  = " $username ($team)";
			   $string .= sprintf("%${length_key}s", ":");
                           $string .= " ";
			   $string .= sprintf("%${length_value}s", $account{$key});
			   $string .= " times";
                           $string .= " ";
			   $string .= sprintf("%6s", $percentage);
                           $string .= "%";
			   $string .= "\n";

			print_date($string);
			
			$string_length  = length($string);
		}
		


		my $key          = "Total :";
		my $value        = $#{$logfile{selected_accounts}} + 1;
		my $length_key   = $longest_key + 2;
		my $length_value = $longest_value - length($value);

		my $string  = " ";
		   $string .= sprintf("%${length_key}s", $key);
		   $string .= " ";
		   $string .= sprintf("%${length_value}s", $value);
		   $string .= " selections";
		   $string .= "\n";

		my $line  = " ";
		   $line .= "-" x ($string_length - 2);
	           $line .= " +\n";

		print_date($line);
		print_date($string);
	}
}

sub parse_logfile {
	my $file = shift;
	my %log  = @_;
	
	print_date("Parsing logfile: $file ...\n") if($cfg{verbose});
	if(-r $file) {
		my $data = '';

		open(F, $file) || die "Error: Cannot open file: $file ($!)";
		while(<F>) {
			$data .= $_;
		}
		close F;

		# [2007-10-14 07:59:04] Random account: 2
		# [2007-10-14 07:59:04]       Username: [DPC]_Fatal_Error_Group0JMT
		# [2007-10-14 07:59:04]           Team: 92
		while($data =~ /\[\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\] Random account: (\d+)\s+\[\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\]       Username: (.*?)\s+\[\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\]           Team: (\d+)\s+/) {
			my $account  = $1;
			my $username = $2;
			my $team     = $3;
			
			$data = $';
			
			if($cfg{verbose}) {
				print_date("Found random account: $account\n");
				print_date("            username: $username\n");
				print_date("                team: $team\n");
			}

			push @{$log{selected_accounts}}, { account  => $account,
			                                   username => $username,
			                                   team     => $team,
							 }
		}

		return %log;
	}
	else {
		print_date("Error: Cannot read file: $file\n");
		exit 1;
	}
}

sub print_usage {
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --directory <PATH>    FAH client working directory\n";
	print "-c, --client <FILE>       FAH client binary to run\n";
	print "-a, --arguments <STRING>  FAH client arguments\n";
	print "-r, --report              Report account statistics\n";
	print "-v, --verbose             Enable verbose output\n";
	print "-h, --help                Show this usage information\n";
}

